from __future__ import annotations

import pickle
import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Literal,
    Mapping,
    Sequence,
    overload,
)

import ase
import lmdb
import numpy as np
from ase import Atoms
from ase.data import chemical_symbols
from typing_extensions import Self, override
from yaml import dump

from .database import DatabaseEntry
from .utils import (
    LazyMapping,
    choose_n,
    freeze_dict,
    intersect,
    k_fold_split,
    random_split,
    split_keeping_ratio,
)


class AtomsDataset(ABC, Sequence[Atoms]):
    """
    An abstract base class for datasets of :class:`ase.Atoms` objects.

    This class provides a common interface for interacting with datasets of
    atomic structures, abstracting over the underlying storage mechanism.

    The two current concrete implementations are
    :class:`~load_atoms.atoms_dataset.InMemoryAtomsDataset`
    and :class:`~load_atoms.atoms_dataset.LmdbAtomsDataset`.
    """

    def __init__(self, description: DatabaseEntry | None = None):
        self.description = description

    @property
    @abstractmethod
    def structure_sizes(self) -> np.ndarray:
        """
        An array containing the number of atoms in each structure, such that:

        .. code-block:: python

            for idx, structure in enumerate(dataset):
                assert len(structure) == dataset.structure_sizes[idx]
        """

    @abstractmethod
    def __len__(self) -> int:
        """The number of structures in the dataset."""

    @overload
    def __getitem__(self, index: int) -> Atoms:
        ...

    @overload
    def __getitem__(
        self, index: list[int] | list[bool] | np.ndarray | slice
    ) -> Self:
        ...

    def __getitem__(  # type: ignore
        self,
        index: int | list[int] | np.ndarray | slice,
    ) -> Atoms | Self:
        r"""
        Get the structure(s) at the given index(es).

        If a single :class:`int` is provided, the corresponding structure is
        returned:

        .. code-block:: pycon

            >>> QM7 = load_dataset("QM7")
            >>> QM7[0]
            Atoms(symbols='CH4', pbc=False)

        If a :class:`slice` is provided, a new :class:`AtomsDataset` is returned
        containing the structures in the slice:

        .. code-block:: pycon

            >>> QM7[:5]
            Dataset:
                structures: 5
                atoms: 32
                species:
                    H: 68.75%
                    C: 28.12%
                    O: 3.12%
                properties:
                    per atom: ()
                    per structure: (energy)

        If a :class:`list` or :class:`numpy.ndarray` of :class:`int`\ s is
        provided, a new :class:`AtomsDataset` is returned containing the
        structures at the given indices:

        .. code-block:: pycon

            >>> len(QM7[[0, 2, 4]])
            3

        If a :class:`list` or :class:`numpy.ndarray` of :class:`bool`\ s is
        provided with the same length as the dataset, a new
        :class:`AtomsDataset` is returned containing the structures where the
        boolean is :code:`True`
        (see also :func:`~load_atoms.AtomsDataset.filter_by`):

        .. code-block:: pycon

            >>> bool_idx = [
            ...     len(s) > 10 for s in QM7
            ... ]
            >>> len(QM7[bool_idx]) == sum(bool_idx)
            True

        Parameters
        ----------
        index
            The index(es) to get the structure(s) at.
        """

        if isinstance(index, slice):
            idxs = range(len(self))[index]
            return self._index_subset(idxs)

        if isinstance(index, list):
            if all(isinstance(i, bool) for i in index):
                if len(index) != len(self):
                    raise ValueError(
                        "Boolean index list must be the same length as the "
                        "dataset."
                    )
                return self._index_subset([i for i, b in enumerate(index) if b])
            else:
                return self._index_subset(index)

        if isinstance(index, np.ndarray):
            return self._index_subset(np.arange(len(self))[index])

        index = int(index)
        return self._index_structure(index)

    def __iter__(self) -> Iterator[ase.Atoms]:
        for i in range(len(self)):
            yield self._index_structure(i)

    @abstractmethod
    def _index_structure(self, index: int) -> Atoms:
        """Get the structure at the given index."""

    @abstractmethod
    def _index_subset(self, idxs: Sequence[int]) -> Self:
        """Get a new dataset containing the structures at the given indices."""

    def __repr__(self) -> str:
        name = "Dataset" if not self.description else self.description.name
        per_atom_properties = (
            "("
            + ", ".join(
                sorted(
                    set(self.arrays.keys())
                    - {
                        "numbers",
                        "positions",
                    }
                )
            )
            + ")"
        )
        per_structure_properties = (
            "(" + ", ".join(sorted(self.info.keys())) + ")"
        )
        species_counts = self.species_counts()
        species_percentages = {
            symbol: f"{count / self.n_atoms:0.2%}"
            for symbol, count in sorted(
                species_counts.items(), key=lambda item: item[1], reverse=True
            )
        }

        return dump(
            {
                name: {
                    "structures": f"{len(self):,}",
                    "atoms": f"{self.n_atoms:,}",
                    "species": species_percentages,
                    "properties": {
                        "per atom": per_atom_properties,
                        "per structure": per_structure_properties,
                    },
                }
            },
            sort_keys=False,
            indent=4,
        )

    @property
    @abstractmethod
    def info(self) -> Mapping[str, Any]:
        r"""
        Get a mapping from keys that are shared across all structures'
        ``.info`` attributes to the concatenated corresponding values.

        The returned mapping conforms to:

        .. code-block:: python

            for key, value in dataset.info.items():
                for i, structure in enumerate(dataset):
                    assert structure.info[key] == value[i]
        """

    @property
    @abstractmethod
    def arrays(self) -> Mapping[str, np.ndarray]:
        """
        Get a mapping from each structure's :code:`.arrays` keys to arrays.

        The returned mapping conforms to:

        .. code-block:: python

            for key, value in dataset.arrays.items():
                assert value.shape[0] == dataset.n_atoms
                assert value == np.vstack(
                    [structure.arrays[key] for structure in dataset]
                )
        """

    @classmethod
    @abstractmethod
    def save(
        cls,
        path: Path,
        structures: Iterable[Atoms],
        description: DatabaseEntry | None = None,
    ):
        """
        Save the dataset to a file.

        Parameters
        ----------
        path
            The path to save the dataset to.
        structures
            The structures to save to the dataset.
        description
            The description of the dataset.
        """

    @classmethod
    @abstractmethod
    def load(cls, path: Path) -> Self:
        """
        Load the dataset from a file.

        Parameters
        ----------
        path
            The path to load the dataset from.
        """
        pass

    # concrete methods

    def species_counts(self) -> Mapping[str, int]:
        """
        Get the number of atoms of each species in the dataset.
        """
        return {
            chemical_symbols[species]: (self.arrays["numbers"] == species).sum()
            for species in np.unique(self.arrays["numbers"])
        }

    def __contains__(self, item: Any) -> bool:
        """
        Check if the dataset contains a structure.

        Warning: this method is not efficient for large datasets.
        """
        return any(item == other for other in self)

    @property
    def n_atoms(self) -> int:
        r"""
        The total number of atoms in the dataset.

        This is equivalent to the sum of the number of atoms in each structure.
        """
        return int(self.structure_sizes.sum())

    def filter_by(
        self,
        *functions: Callable[[ase.Atoms], bool],
        **info_kwargs: Any,
    ) -> Self:
        """
        Return a new dataset containing only the structures that match the
        given criteria.

        Parameters
        ----------
        functions
            Functions to filter the dataset by. Each function should take an
            ASE Atoms object as input and return a boolean.
        info_kwargs
            Keyword arguments to filter the dataset by. Only atoms objects with
            matching info keys and values will be returned.

        Example
        -------
        Get small, amorphous structures with large forces:

        .. code-block:: pycon
            :emphasize-lines: 3-7

            >>> from load_atoms import load_dataset
            >>> dataset = load_dataset("C-GAP-17")
            >>> dataset.filter_by(
            ...     lambda structure: len(structure) < 50,
            ...     lambda structure: structure.arrays["force"].max() > 5,
            ...     config_type="bulk_amo"
            ... )
            Dataset:
                structures: 609
                atoms: 23,169
                species:
                    C: 100.00%
                properties:
                    per atom: (force)
                    per structure: (config_type, detailed_ct, split, energy)
        """

        def matches_info(structure: ase.Atoms) -> bool:
            for key, value in info_kwargs.items():
                if structure.info.get(key, None) != value:
                    return False
            return True

        functions = (*functions, matches_info)

        def the_filter(structure: ase.Atoms) -> bool:
            return all(f(structure) for f in functions)

        index = [i for i, structure in enumerate(self) if the_filter(structure)]
        return self[index]

    def random_split(
        self,
        splits: Sequence[float] | Sequence[int],
        seed: int = 42,
        keep_ratio: str | None = None,
    ) -> list[Self]:
        r"""
        Randomly split the dataset into multiple, disjoint parts.

        Parameters
        ----------
        splits
            The number of structures to put in each split.
            If a list of :class:`float`\ s, the splits will be
            calculated as a fraction of the dataset size.
        seed
            The random seed to use for shuffling the dataset.
        keep_ratio
            If not :code:`None`, splits will be generated to maintain the
            ratio of structures in each split with the specified :code:`.info`
            value.

        Returns
        -------
        list[Self]
            A list of new datasets, each containing a subset of the original

        Examples
        --------
        Split a :code:`dataset` into 80% training and 20% test sets:

        >>> train, test = dataset.random_split([0.8, 0.2])

        Split a :code:`dataset` into 3 parts:

        >>> train, val, test = dataset.random_split([1_000, 100, 100])

        Maintain the ratio of :code:`config_type` values in each split:

        .. code-block:: pycon
            :emphasize-lines: 16-19

            >>> from load_atoms import load_dataset
            >>> import numpy as np
            >>> # helper function
            >>> def ratios(thing):
            ...     values, counts = np.unique(thing, return_counts=True)
            ...     max_len = max(len(str(v)) for v in values)
            ...     for v, c in zip(values, counts / counts.sum()):
            ...         print(f"{v:>{max_len}}: {c:>6.2%}")
            ...
            >>> dataset = load_dataset("C-GAP-17")
            >>> ratios(dataset.info["config_type"])
              bulk_amo: 75.28%
            bulk_cryst:  8.83%
                 dimer:  0.66%
              surf_amo: 15.23%
            >>> train, val, test = dataset.random_split(
            ...     [0.6, 0.2, 0.2],
            ...     keep_ratio="config_type"
            ... )
            >>> ratios(train.info["config_type"])
              bulk_amo: 75.28%
            bulk_cryst:  8.83%
                 dimer:  0.66%
              surf_amo: 15.23%
        """

        if keep_ratio is None:
            return [
                self[split]
                for split in random_split(range(len(self)), splits, seed)
            ]

        if keep_ratio not in self.info:
            raise KeyError(
                f"Unknown key {keep_ratio}. "
                "Available keys are: " + ", ".join(self.info.keys())
            )

        if isinstance(splits[0], int):
            final_sizes: list[int] = splits  # type: ignore
        else:
            final_sizes = [int(s * len(self)) for s in splits]

        normalised_fractional_splits = [s / sum(splits) for s in splits]

        split_idxs = split_keeping_ratio(
            range(len(self)),
            group_ids=self.info[keep_ratio],
            splitting_function=partial(
                random_split, seed=seed, splits=normalised_fractional_splits
            ),
        )

        return [
            self[choose_n(split, size, seed)]
            for split, size in zip(split_idxs, final_sizes)
        ]

    def k_fold_split(
        self,
        k: int = 5,
        fold: int = 0,
        shuffle: bool = True,
        seed: int = 42,
        keep_ratio: str | None = None,
    ) -> tuple[Self, Self]:
        """
        Generate (an optionally shuffled) train/test split for cross-validation.

        Parameters
        ----------
        k
            The number of folds to use.
        fold
            The fold to use for testing.
        shuffle
            Whether to shuffle the dataset before splitting.
        seed
            The random seed to use for shuffling the dataset.
        keep_ratio
            If not :code:`None`, splits will be generated to maintain the
            ratio of structures in each split with the specified :code:`.info`
            value.

        Returns
        -------
        Tuple[Self, Self]
            The train and test datasets.

        Example
        -------

        Basic usage:

        .. code-block:: pycon
            :emphasize-lines: 2

            >>> for i in range(5):
            ...     train, test = dataset.k_fold_split(k=5, fold=i)
            ...     ...  # do something, e.g. train a model

        Maintain the ratio of :code:`config_type` values in each split
        (see also :func:`~load_atoms.AtomsDataset.random_split` for a more
        detailed example of this feature):

        .. code-block:: pycon

            >>> train, test = dataset.k_fold_split(
            ...     k=5, fold=0, keep_ratio="config_type"
            ... )

        """

        if k < 2:
            raise ValueError("k must be at least 2")

        fold = fold % k

        if shuffle:
            idxs = np.random.RandomState(seed).permutation(len(self))
        else:
            idxs = np.arange(len(self))

        if keep_ratio is None:
            train_idxs, test_idxs = k_fold_split(idxs.tolist(), k, fold)
        else:
            if keep_ratio not in self.info:
                raise KeyError(
                    f"Unknown key {keep_ratio}. "
                    "Available keys are: " + ", ".join(self.info.keys())
                )
            if not shuffle:
                raise ValueError(
                    "Keep ratio splits are only supported when shuffling."
                )

            group_ids = self.info[keep_ratio][idxs]
            train_idxs, test_idxs = split_keeping_ratio(
                idxs.tolist(), group_ids, partial(k_fold_split, k=k, fold=fold)
            )

        return self[train_idxs], self[test_idxs]


class InMemoryAtomsDataset(AtomsDataset):
    """
    An in-memory implementation of :class:`AtomsDataset`.

    Internally, this class wraps a :class:`list` of :class:`ase.Atoms` objects,
    all of which are stored in RAM. Suitable for small to moderately large
    datasets.
    """

    def __init__(
        self,
        structures: list[ase.Atoms],
        description: DatabaseEntry | None = None,
    ):
        super().__init__(description)

        if len(structures) == 1:
            warnings.warn(
                "Creating a dataset with a single structure. "
                "Typically, datasets contain multiple structures - "
                "did you mean to do this?",
                stacklevel=2,
            )

        self._structures = structures
        self._info = _get_info_mapping(structures)
        self._arrays = _get_arrays_mapping(structures)

    @property
    @override
    def info(self) -> LazyMapping[str, Any]:
        return self._info

    @property
    @override
    def arrays(self) -> LazyMapping[str, np.ndarray]:
        return self._arrays

    @property
    @override
    def structure_sizes(self) -> np.ndarray:
        return np.array([len(s) for s in self._structures])

    @override
    def __len__(self) -> int:
        return len(self._structures)

    @override
    def _index_structure(self, index: int) -> Atoms:
        return self._structures[index]

    @override
    def _index_subset(self, idxs: Sequence[int]) -> InMemoryAtomsDataset:
        return InMemoryAtomsDataset([self._index_structure(i) for i in idxs])

    @override
    @classmethod
    def save(
        cls,
        path: Path,
        structures: Iterable[Atoms],
        description: DatabaseEntry | None = None,
    ):
        path.parent.mkdir(parents=True, exist_ok=True)
        to_save = {
            "structures": list(structures),
            "description": description,
        }
        with open(path, "wb") as f:
            pickle.dump(to_save, f)

    @classmethod
    @override
    def load(cls, path: Path) -> InMemoryAtomsDataset:
        with open(path, "rb") as f:
            data = pickle.load(f)
        return cls(**data)


@dataclass
class LmdbMetadata:
    structure_sizes: np.ndarray
    species_per_structure: list[dict[str, int]]
    per_atom_properties: list[str]
    per_structure_properties: list[str]


class LmdbAtomsDataset(AtomsDataset):
    r"""
    An LMDB-backed implementation of :class:`AtomsDataset`.

    Internally, this class wraps an :class:`lmdb.Environment` object, which
    stores the dataset in an LMDB database. Suitable for large datasets that
    cannot fit in memory. Accessing data from this dataset type is (marginally)
    slower than for :class:`InMemoryAtomsDataset`\ s, but allows for efficient
    processing of extremely large datasets that cannot otherwise fit in memory.

    .. warning::

        The :class:`ase.Atoms` objects in an LMDB dataset are read-only.
        Modifying the :code:`.info` or :code:`.arrays` of an :class:`ase.Atoms`
        object will have no effect, and will instead throw an error.
    """

    def __init__(self, path: Path, idx_subset: np.ndarray | None = None):
        self.path = path

        # setup lmdb environment, and keep a transaction open for the lifetime
        # of the dataset (to enable fast reads)
        self.env = lmdb.open(
            str(path), readonly=True, lock=False, map_async=True
        )
        self.txn = self.env.begin(write=False)

        super().__init__(
            description=pickle.loads(
                self.txn.get("description".encode("ascii"))
            )
        )

        self.metadata: LmdbMetadata = pickle.loads(
            self.txn.get("metadata".encode("ascii"))
        )
        self.idx_subset = (
            idx_subset
            if idx_subset is not None
            else np.arange(len(self.metadata.structure_sizes))
        )

        # TODO: add warnings to loaders about potential slowness
        self._info = _get_info_mapping(
            structures=self,
            keys=self.metadata.per_structure_properties,
            loader_warning=(
                "LmdbAtomsDatasets do not hold all structure properties in "
                "memory: accessing .info will be slow."
            ),
        )
        self._arrays = _get_arrays_mapping(
            structures=self,
            keys=self.metadata.per_atom_properties + ["numbers", "positions"],
            loader_warning=(
                "LmdbAtomsDatasets do not hold all per-atom properties in "
                "memory: accessing .arrays will be slow."
            ),
        )

    def close(self):
        self.txn.close()
        self.env.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _index_structure(self, index: int) -> Atoms:
        index = self.idx_subset[index]
        pickled_data = self.txn.get(f"{index}".encode("ascii"))
        atoms = pickle.loads(pickled_data)
        error_msg = (
            "The Atoms objects in an LMDB dataset are read-only: "
            "modifying the {} of an Atoms object will have no effect."
        )
        atoms.info = freeze_dict(atoms.info, error_msg.format("info"))
        atoms.arrays = freeze_dict(atoms.arrays, error_msg.format("arrays"))
        return atoms

    @property
    @override
    def structure_sizes(self) -> np.ndarray:
        sizes = self.metadata.structure_sizes
        if self.idx_subset is not None:
            return sizes[self.idx_subset]
        return sizes

    @override
    def __len__(self) -> int:
        return len(self.structure_sizes)

    @override
    def species_counts(self) -> Mapping[str, int]:
        to_sum: list[dict[str, int]] = []
        all_species: set[str] = set()

        for idx in self.idx_subset:
            species_count = self.metadata.species_per_structure[idx]
            to_sum.append(species_count)
            all_species.update(species_count.keys())

        summed = {symbol: 0 for symbol in all_species}
        for species_count in to_sum:
            for symbol, count in species_count.items():
                summed[symbol] += count
        return summed

    @override
    def __contains__(self, item: Any) -> bool:
        warnings.warn(
            "Checking if an LMDB dataset contains a structure is slow "
            "because it requires loading every structure into memory. "
            "Consider using a different dataset format if possible.",
            stacklevel=2,
        )
        return any(item == structure for structure in self)

    @override
    def _index_subset(self, idxs: Sequence[int]) -> LmdbAtomsDataset:
        return LmdbAtomsDataset(self.path, self.idx_subset[idxs])

    @property
    @override
    def info(self) -> LazyMapping[str, Any]:
        return self._info

    @property
    @override
    def arrays(self) -> LazyMapping[str, np.ndarray]:
        return self._arrays

    @classmethod
    @override
    def load(cls, path: Path) -> LmdbAtomsDataset:
        return cls(path)

    @classmethod
    @override
    def save(
        cls,
        path: Path,
        structures: Iterable[Atoms],
        description: DatabaseEntry | None = None,
    ):
        path.mkdir(parents=True, exist_ok=True)

        one_TB = int(1e12)
        env = lmdb.open(str(path), map_size=one_TB)

        with env.begin(write=True) as txn:
            structure_sizes = []
            species_per_structure = []
            per_atom_properties: list[set[str]] = []
            per_structure_properties: list[set[str]] = []

            for idx, structure in enumerate(structures):
                # Save structure
                txn.put(f"{idx}".encode("ascii"), pickle.dumps(structure))

                # Update metadata
                structure_sizes.append(len(structure))
                species_per_structure.append(
                    {
                        chemical_symbols[Z]: (
                            structure.arrays["numbers"] == Z
                        ).sum()
                        for Z in np.unique(structure.arrays["numbers"])
                    }
                )
                per_atom_properties.append(
                    set(structure.arrays.keys()) - {"numbers", "positions"}
                )
                per_structure_properties.append(set(structure.info.keys()))

            # Save metadata
            metadata = LmdbMetadata(
                structure_sizes=np.array(structure_sizes),
                species_per_structure=species_per_structure,
                per_atom_properties=sorted(intersect(per_atom_properties)),
                per_structure_properties=sorted(
                    intersect(per_structure_properties)
                ),
            )
            txn.put("metadata".encode("ascii"), pickle.dumps(metadata))

            # Save description if provided
            if description:
                txn.put(
                    "description".encode("ascii"), pickle.dumps(description)
                )

        env.close()


def _get_info_mapping(
    structures: Iterable[Atoms],
    keys: list[str] | None = None,
    loader_warning: str | None = None,
) -> LazyMapping[str, np.ndarray]:
    if keys is None:
        keys = list(intersect(s.info.keys() for s in structures))

    def loader(key: str):
        if loader_warning is not None:
            warnings.warn(loader_warning, stacklevel=2)
        return np.array([s.info[key] for s in structures])

    return LazyMapping(keys, loader)


def _get_arrays_mapping(
    structures: Iterable[Atoms],
    keys: list[str] | None = None,
    loader_warning: str | None = None,
) -> LazyMapping[str, np.ndarray]:
    if keys is None:
        keys = list(intersect(s.arrays.keys() for s in structures))

    def loader(key: str):
        if loader_warning is not None:
            warnings.warn(loader_warning, stacklevel=2)
        return np.concatenate([s.arrays[key] for s in structures])

    return LazyMapping(keys, loader)


def summarise_dataset(
    structures: list[Atoms] | AtomsDataset,
    description: DatabaseEntry | None = None,
) -> str:
    if isinstance(structures, AtomsDataset):
        return str(structures)
    return str(InMemoryAtomsDataset(structures, description))


def get_file_extension_and_dataset_class(
    format: Literal["lmdb", "memory"]
) -> tuple[str, type[AtomsDataset]]:
    return {
        "lmdb": ("lmdb", LmdbAtomsDataset),
        "memory": ("pkl", InMemoryAtomsDataset),
    }[format]



import os

from load_atoms import load_dataset
from setup import TESTING_DIR


def test_verbose(capfd):
    # Test verbose output
    os.environ["LOAD_ATOMS_VERBOSE"] = "1"
    captured = capfd.readouterr()
    load_dataset("C-GAP-17", root=TESTING_DIR)
    captured = capfd.readouterr()
    assert "C-GAP-17" in captured.out

    # Test non-verbose output
    os.environ["LOAD_ATOMS_VERBOSE"] = "0"
    load_dataset("C-GAP-17", root=TESTING_DIR)
    captured = capfd.readouterr()
    assert "C-GAP-17" not in captured.out



from __future__ import annotations

from pathlib import Path
from typing import Iterator

import ase.io
from ase import Atoms
from load_atoms.database.backend import BaseImporter, rename, unzip_file
from load_atoms.database.internet import FileDownload
from load_atoms.progress import Progress


class Importer(BaseImporter):
    @classmethod
    def files_to_download(cls) -> list[FileDownload]:
        return [
            FileDownload(
                url="https://zenodo.org/records/7704087/files/jla-gardner/carbon-data-v1.0.zip",
                expected_hash="b43fc702ef6d",
            )
        ]

    @classmethod
    def get_structures(
        cls, tmp_dir: Path, progress: Progress
    ) -> Iterator[Atoms]:
        # Unzip the file
        contents_path = unzip_file(tmp_dir / "carbon-data-v1.0.zip", progress)

        extxyz_files = sorted(contents_path.glob("**/*.extxyz"))
        task = progress.new_task(
            f"Processing {len(extxyz_files)} .extxyz files",
            total=len(extxyz_files),
        )

        # iterate through all .extxyz files
        for file_path in extxyz_files:
            structures = ase.io.read(file_path, index=":")
            assert isinstance(structures, list)
            for structure in structures:
                yield process_structure(structure)

            task.update(advance=1)


def process_structure(structure: Atoms) -> Atoms:
    structure = rename(
        structure,
        {
            "gap17_forces": "forces",
            "gap17_energy": "local_energies",
        },
    )
    structure.info["energy"] = structure.arrays["local_energies"].sum()
    return structure



from __future__ import annotations

from pathlib import Path
from typing import Iterator

import ase.io
from ase import Atoms
from load_atoms.database.backend import BaseImporter, rename, unzip_file
from load_atoms.database.internet import FileDownload
from load_atoms.progress import Progress


class Importer(BaseImporter):
    @classmethod
    def files_to_download(cls) -> list[FileDownload]:
        return [
            FileDownload(
                url="https://zenodo.org/records/6353684/files/sio2_potential_data.zip",
                expected_hash="98ea6e58f6d9",
            )
        ]

    @classmethod
    def get_structures(
        cls, tmp_dir: Path, progress: Progress
    ) -> Iterator[Atoms]:
        contents = unzip_file(tmp_dir / "sio2_potential_data.zip", progress)

        for structure in ase.io.iread(
            contents / "sio2_potential_data/database/dataset.scan.2.xyz"
        ):
            yield rename(structure, {"virials": "virial"})



from __future__ import annotations

from pathlib import Path
from typing import Iterator

import ase.io
from ase import Atoms
from load_atoms.database.backend import BaseImporter, rename, unzip_file
from load_atoms.database.internet import FileDownload
from load_atoms.progress import Progress


class Importer(BaseImporter):
    @classmethod
    def files_to_download(cls) -> list[FileDownload]:
        return [
            FileDownload(
                url="https://zenodo.org/record/1250555/files/libAtoms/silicon-testing-framework-v1.0.zip",
                expected_hash="da0462802df1",
                local_name="zip-file.zip",
            )
        ]

    @classmethod
    def get_structures(
        cls, tmp_dir: Path, progress: Progress
    ) -> Iterator[Atoms]:
        contents_path = unzip_file(tmp_dir / "zip-file.zip", progress)

        for structure in ase.io.iread(
            contents_path
            / "libAtoms-silicon-testing-framework-fc252cb/models/GAP/gp_iter6_sparse9k.xml.xyz"  # noqa: E501
        ):
            yield rename(
                structure,
                {
                    "DFT_force": "forces",
                    "dft_force": "forces",
                    "DFT_energy": "energy",
                    "dft_energy": "energy",
                },
            )



# testing whether a visualisation works is difficult
# as a first stab, we can make sure that the function runs,
# and that the utility functions are working as expected

import pytest
from ase.build import molecule
from load_atoms import view
from load_atoms.visualisation import unique_variable_name


@pytest.mark.parametrize("show_bonds", [True, False])
def test_view(show_bonds):
    ethanol = molecule("CH3CH2OH")
    view(ethanol, show_bonds)

    ethanol.center(vacuum=5)
    ethanol.pbc = [True, True, True]
    view(ethanol, show_bonds)


def test_unique_variable_name():
    # ensure that the function returns a unique name
    # that is also a valid javascript variable name
    names = [unique_variable_name() for _ in range(10)]
    assert len(names) == len(set(names)), "Names should be unique"

    for name in names:
        assert name.isidentifier(), "Name should be a valid identifier"
        assert name[0].isalpha(), "Name should start with a letter"
        assert all(
            c.isalnum() or c == "_" for c in name
        ), "Name should only contain letters, numbers, or underscores"



from load_atoms.database.backend import BASE_GITHUB_URL, SingleFileImporter
from load_atoms.database.internet import FileDownload


class Importer(SingleFileImporter):
    @classmethod
    def file_to_download(cls) -> FileDownload:
        return FileDownload(
            url=f"{BASE_GITHUB_URL}/C-GAP-20U/C-GAP-20U.xyz",
            expected_hash="da0462802df1",
        )



from load_atoms.database.backend import BASE_GITHUB_URL, SingleFileImporter
from load_atoms.database.internet import FileDownload


class Importer(SingleFileImporter):
    @classmethod
    def file_to_download(cls) -> FileDownload:
        return FileDownload(
            url=f"{BASE_GITHUB_URL}/GST-GAP-22/refitted_GST-GAP-22_PBE.xyz",
            expected_hash="e4c467026dc0",
        )



import argparse
from pathlib import Path

import ase.io
from load_atoms import load_dataset


def main():
    parser = argparse.ArgumentParser(
        description=(
            "Download a load_atoms dataset from the internet to "
            "a local, `ase.io.read`able file."
        )
    )
    parser.add_argument("dataset_id", help="ID of the dataset to download")
    parser.add_argument(
        "--format",
        default="xyz",
        required=False,
        help="""\
Format to save the dataset in. Must be one of the formats supported by
`ase.io.write`.
""",
    )
    parser.add_argument(
        "--root",
        default=".",
        help="Root directory to save the dataset (default: current directory)",
        required=False,
    )

    args = parser.parse_args()

    # load the dataset in the usual way
    dataset = load_dataset(args.dataset_id)

    if len(dataset) > 10_000:
        print(
            f"""\
warning: {args.dataset_id} has {len(dataset)} structures. 
         Are you sure you want to save this to disk?
         (y/n) """
        )
        response = input()
        if not response.lower().startswith("y"):
            print("Exiting...")
            return

    file_to_write_to = Path(args.root) / f"{args.dataset_id}.{args.format}"
    file_to_write_to.parent.mkdir(parents=True, exist_ok=True)
    ase.io.write(file_to_write_to, dataset)

    print(f"Dataset saved to {file_to_write_to}")



from __future__ import annotations

import warnings
from dataclasses import dataclass
from pathlib import Path

import requests

from load_atoms.progress import Progress
from load_atoms.utils import matches_checksum


@dataclass
class FileDownload:
    url: str
    expected_hash: str
    local_name: str = None  # type: ignore

    def __post_init__(self):
        if self.local_name is None:
            self.local_name = Path(self.url).name


def download(url: str, local_path: Path, progress: Progress):
    """
    Download a file from the given url to the given path.

    If path is a file, the file will be downloaded to that path.
    Else, the file will be downloaded to the given path, with the same name as
    the file at the given url.

    Parameters
    ----------
    url
        The url to download the file from.
    local_path
        The path to download the file to.
    progress
        The progress bar to add the download to.
    """

    if local_path.is_dir():
        local_path = local_path / Path(url).name
    local_path.parent.mkdir(parents=True, exist_ok=True)

    with requests.get(url, stream=True) as response, progress.new_task(
        f"Downloading {local_path.name}"
    ) as task, open(local_path, "wb") as f:
        response.raise_for_status()
        file_size = int(response.headers["content-length"])
        task.update(total=file_size)
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)
                task.update(advance=len(chunk))


def download_all(
    files: list[FileDownload],
    download_dir: Path,
    progress: Progress,
):
    """
    Download all files and verify their checksums.

    Parameters
    ----------
    files
        A list of FileDownloads
    tmp_dir
        The directory where the files should be saved.
    progress
        A Progress object to track the download progress.
    """

    # 1. download any missing files
    for file in files:
        local_path = download_dir / file.local_name
        if not local_path.exists():
            download(file.url, local_path, progress)

    # 2. verify the hashes of all files
    for file in files:
        local_path = download_dir / file.local_name
        if not matches_checksum(local_path, file.expected_hash):
            warnings.warn(
                f"Checksum mismatch for file: {local_path}",
                stacklevel=2,
            )



# explicitly not using future annotations since this is not supported
# by pydantic for python versions we want to target

from pathlib import Path
from typing import Dict, Literal, Optional, Union

import yaml
from pydantic import BaseModel, field_validator

from load_atoms.utils import BASE_REMOTE_URL

LICENSE_URLS = {
    "CC BY-NC-SA 4.0": "https://creativecommons.org/licenses/by-nc-sa/4.0/deed.en",
    "CC BY-NC 4.0": "https://creativecommons.org/licenses/by-nc/4.0/deed.en",
    "CC BY 4.0": "https://creativecommons.org/licenses/by/4.0/deed.en",
    "CC0": "https://creativecommons.org/publicdomain/zero/1.0/",
    "MIT": "https://opensource.org/licenses/MIT",
    "GPLv3": "https://www.gnu.org/licenses/gpl-3.0.html",
}
VALID_LICENSES = list(LICENSE_URLS.keys())

VALID_CATEGORIES = ["Benchmarks", "Potential Fitting", "Synthetic Data"]


class PropertyDescription(BaseModel):
    """
    Holds a description of a property, such that it can be automatically
    validated upon creation.
    """

    desc: str
    """A description of the property"""

    units: Optional[str] = None
    """The units of the property"""


class DatabaseEntry(BaseModel):
    """
    Holds all the required metadata for a named dataset, such that it can be
    automatically downloaded using :func:`~load_atoms.load_dataset`, and so that
    documentation can be automatically generated.
    """

    name: str
    """The name of the dataset"""

    year: int
    """The year the dataset was created"""

    description: str
    """A description of the dataset (in ``.rst`` format)"""

    category: str
    """
    The category of the dataset 
    (e.g. ``"Potential Fitting"``, ``"Benchmarks"``)
    """

    format: Literal["lmdb", "memory"] = "memory"
    """The format of the dataset"""

    minimum_load_atoms_version: Union[str, None] = None
    """
    The minimum version of load-atoms that is required to load the dataset.
    """

    citation: Optional[str] = None
    """A citation for the dataset (in BibTeX format)"""

    license: Optional[str] = None
    """The license identifier of the dataset (e.g. ``"CC BY-NC-SA 4.0"``)"""

    representative_structure: Optional[int] = None
    """The index of a representative structure (for visualisation purposes)"""

    per_atom_properties: Optional[Dict[str, PropertyDescription]] = None
    """A mapping from per-atom properties to their descriptions"""

    per_structure_properties: Optional[Dict[str, PropertyDescription]] = None
    """A mapping from per-structure properties to their descriptions"""

    @field_validator("category")
    def validate_category(cls, v):
        if v not in VALID_CATEGORIES:
            raise ValueError(
                f"Invalid category: {v}. Must be one of {VALID_CATEGORIES}"
            )
        return v

    @field_validator("license")
    def validate_license(cls, v):
        if v not in VALID_LICENSES:
            raise ValueError(
                f"Invalid license: {v}. Must be one of {VALID_LICENSES}"
            )
        return v

    @field_validator("citation")
    def validate_citation(cls, v):
        v = v.strip()
        if v.startswith("@") and v.endswith("}"):
            return v
        raise ValueError(f"Invalid BibTeX: {v}")

    @field_validator("minimum_load_atoms_version", mode="before")
    def convert_minimum_version_to_str(cls, v):
        if v is None:
            return None
        return str(v)

    @classmethod
    def from_yaml_file(cls, path: Union[Path, str]) -> "DatabaseEntry":
        path = Path(path).resolve()
        with open(path) as f:
            data = yaml.safe_load(f)

        try:
            return cls(**data)
        except Exception as e:
            raise ValueError(
                f"Error loading dataset description from {path}. It may be "
                "that you have a stale version of this dataset's yaml file on "
                "disk. Please delete the file and try again:\n"
                f'   $ rm "{path}"'
            ) from e

    @classmethod
    def remote_url_for_yaml(cls, dataset_id: str) -> str:
        return BASE_REMOTE_URL + f"{dataset_id}/{dataset_id}.yaml"

    @classmethod
    def importer_file_stem(cls, dataset_id: str) -> str:
        return dataset_id.lower().replace("-", "_")

    @classmethod
    def remote_url_for_importer(cls, dataset_id: str) -> str:
        fname = DatabaseEntry.importer_file_stem(dataset_id)
        return BASE_REMOTE_URL + f"src/load_atoms/database/importers/{fname}.py"



from load_atoms.database.backend import SingleFileImporter
from load_atoms.database.internet import FileDownload


class Importer(SingleFileImporter):
    @classmethod
    def file_to_download(cls) -> FileDownload:
        return FileDownload(
            url="https://zenodo.org/record/4003703/files/P_GAP_20_fitting_data.xyz",
            expected_hash="ab3059018068",
        )



import shutil
from contextlib import nullcontext
from pathlib import Path

import numpy as np
import pytest
from ase import Atoms
from ase.io import read, write
from load_atoms import load_dataset
from load_atoms.atoms_dataset import (
    AtomsDataset,
    LmdbAtomsDataset,
    summarise_dataset,
)
from load_atoms.utils import UnknownDatasetException
from setup import TESTING_DIR

STRUCTURES = [Atoms("H2O"), Atoms("H2O2")]
GAP17 = load_dataset("C-GAP-17", root=TESTING_DIR)

shutil.rmtree(TESTING_DIR / "gap17.lmdb", ignore_errors=True)
LmdbAtomsDataset.save(
    path=TESTING_DIR / "gap17.lmdb",
    structures=GAP17,
    description=GAP17.description,
)
GAP17_LMDB = LmdbAtomsDataset(TESTING_DIR / "gap17.lmdb")


def _is_water_dataset(dataset):
    assert len(dataset) == 2
    assert set(dataset[0].symbols) == {"H", "O"}


def test_dataset_from_structures():
    dataset = load_dataset(STRUCTURES)
    _is_water_dataset(dataset)


def test_dataset_writeable_and_readable(tmp_path):
    dataset = load_dataset(STRUCTURES)
    write(tmp_path / "test.xyz", dataset)

    structures = read(tmp_path / "test.xyz", index=":")
    assert isinstance(structures, list)
    dataset2 = load_dataset(structures)
    _is_water_dataset(dataset2)

    dataset3 = load_dataset(tmp_path / "test.xyz")
    _is_water_dataset(dataset3)


@pytest.mark.filterwarnings("ignore:Creating a dataset with a single structure")
def test_indexing():
    dataset = load_dataset(STRUCTURES)

    structure = dataset[0]
    assert isinstance(
        structure, Atoms
    ), "Indexing should return an Atoms object"
    assert (
        structure is STRUCTURES[0]
    ), "Indexing should return the same structure"

    sub_dataset = dataset[1:]
    assert isinstance(
        sub_dataset, AtomsDataset
    ), "Slicing should return a dataset"
    assert (
        len(sub_dataset) == 1
    ), "Slicing should return the correct number of structures"

    indices = np.array([0, 1, 0, 1])
    sub_dataset = dataset[indices]
    assert isinstance(
        sub_dataset, AtomsDataset
    ), "Indexing should return a dataset"
    assert (
        len(sub_dataset) == 4
    ), "Indexing should return the correct number of structures"

    indices = np.array([True, False])
    sub_dataset = dataset[indices]
    assert isinstance(
        sub_dataset, AtomsDataset
    ), "Indexing should return a dataset"
    assert (
        len(sub_dataset) == 1
    ), "Indexing should return the correct number of structures"

    indices = [0, 1, 1]
    sub_dataset = dataset[indices]
    assert isinstance(
        sub_dataset, AtomsDataset
    ), "Indexing should return a dataset"
    assert (
        len(sub_dataset) == 3
    ), "Indexing should return the correct number of structures"


@pytest.mark.parametrize(
    "dataset", [GAP17, GAP17_LMDB], ids=["gap17", "gap17_lmdb"]
)
def test_can_load_from_id(tmp_path, dataset):
    assert len(dataset) == 4530

    with pytest.raises(UnknownDatasetException):
        load_dataset("made_up_dataset", root=tmp_path)


@pytest.mark.parametrize(
    "dataset", [GAP17, GAP17_LMDB], ids=["gap17", "gap17_lmdb"]
)
def test_summarise(dataset):
    summary = summarise_dataset(STRUCTURES)
    assert summary.startswith(
        "Dataset"
    ), "Unknown datasets should start like this"

    summary = repr(dataset)
    assert "energy" in summary, "The summary should contain the property names"
    assert summary.startswith(
        "C-GAP-17"
    ), "The summary should contain the dataset name"


def test_appears_equal():
    assert str(GAP17) == str(GAP17_LMDB)


def test_useful_error_message():
    with pytest.raises(
        TypeError,
        match="provide a string, a list of structures, or a path to a file.",
    ):
        load_dataset(1)  # type: ignore

    with pytest.raises(
        ValueError,
        match="The provided path does not exist.",
    ):
        load_dataset(Path("made_up_file.xyz"))


def test_useful_warning(tmp_path):
    structure = Atoms("H2O")
    write(tmp_path / "test.xyz", structure)

    with pytest.warns(
        UserWarning,
        match="single structure",
    ):
        load_dataset(tmp_path / "test.xyz")


@pytest.mark.parametrize(
    "dataset", [GAP17, GAP17_LMDB], ids=["gap17", "gap17_lmdb"]
)
def test_info_and_arrays(dataset):
    context = nullcontext() if dataset is GAP17 else pytest.warns(UserWarning)
    with context:
        assert "energy" in dataset.info
        assert isinstance(dataset.info["energy"], np.ndarray)

    with context:
        assert "positions" in dataset.arrays
        assert dataset.arrays["positions"].shape[-1] == 3


def test_properties():
    dataset = load_dataset(STRUCTURES)

    assert len(dataset) == 2
    assert dataset.n_atoms == 7
    assert (dataset.structure_sizes == [3, 4]).all()


def _get_proportion(config_type, dataset):
    return np.sum(dataset.info["config_type"] == config_type) / len(dataset)


@pytest.mark.parametrize(
    "dataset", [GAP17, GAP17_LMDB], ids=["gap17", "gap17_lmdb"]
)
def test_random_split(dataset):
    # request integer splits
    train, test = dataset.random_split([100, 50])
    assert len(train) == 100
    assert len(test) == 50

    # request float splits
    n = len(dataset)
    a, b, c = dataset.random_split([0.5, 0.25, 0.25])
    assert len(a) == n // 2
    assert len(b) in [n // 4, n // 4 + 1]
    assert len(c) in [n // 4, n // 4 + 1]

    # test keep ratio
    context = nullcontext() if dataset is GAP17 else pytest.warns(UserWarning)
    with context:
        assert "config_type" in dataset.info
        assert "bulk_amo" in dataset.info["config_type"]

        # ... with floats
        a, b = dataset.random_split([0.5, 0.5], keep_ratio="config_type")
        assert len(a) + len(b) == len(dataset)
        assert np.isclose(
            _get_proportion("bulk_amo", a), _get_proportion("bulk_amo", dataset)
        )

        # ... with ints
        a, b = dataset.random_split([500, 500], keep_ratio="config_type")
        assert np.isclose(
            _get_proportion("bulk_amo", a),
            _get_proportion("bulk_amo", dataset),
            atol=0.01,
        )
        assert len(a) == len(b) == 500

    # test error
    assert "made_up" not in dataset.info
    with pytest.raises(KeyError):
        dataset.random_split([0.5, 0.5], keep_ratio="made_up")


@pytest.mark.parametrize(
    "dataset", [GAP17, GAP17_LMDB], ids=["gap17", "gap17_lmdb"]
)
def test_k_fold_split(dataset):
    # error messages
    with pytest.raises(ValueError, match="k must be at least 2"):
        dataset.k_fold_split(k=1)

    # atoms are not hashable, so we set a unique id for each structure
    def get_id(structure):
        return float(
            structure.info["energy"] + np.abs(structure.arrays["forces"]).sum()
        )

    a, b = dataset.k_fold_split(5, fold=0, shuffle=False)
    assert a[0] == dataset[0]

    a, b = dataset.k_fold_split(5, fold=0)
    assert len(a) == 3624
    assert len(b) == 906
    a_ids = set(map(get_id, a))
    b_ids = set(map(get_id, b))
    assert a_ids & b_ids == set()

    c, d = dataset.k_fold_split(5, fold=1)
    c_ids = set(map(get_id, c))
    # ensure that b is completely within c
    assert b_ids <= c_ids

    # ensure that the folds completely cover the dataset
    all_test_ids = []
    for i in range(5):
        _, test = dataset.k_fold_split(5, fold=i)
        all_test_ids.extend(list(map(get_id, test)))

    assert len(set(all_test_ids)) == len(dataset)

    # test keep ratio
    a, b = dataset.k_fold_split(k=5, fold=0, keep_ratio="config_type")
    context = nullcontext() if dataset is GAP17 else pytest.warns(UserWarning)
    with context:
        assert np.isclose(
            _get_proportion("bulk_amo", a), _get_proportion("bulk_amo", dataset)
        )

    # test error
    assert "made_up" not in dataset.info
    with pytest.raises(KeyError):
        dataset.k_fold_split(5, fold=0, keep_ratio="made_up")

    with pytest.raises(ValueError, match="only supported when shuffling"):
        dataset.k_fold_split(5, fold=0, keep_ratio="config_type", shuffle=False)


def test_read_only():
    dataset = GAP17_LMDB
    atoms = dataset[0]
    with pytest.raises(ValueError, match="info"):
        atoms.info["test"] = 1
    with pytest.raises(ValueError, match="arrays"):
        atoms.arrays["test"] = np.zeros((1, 3))


@pytest.mark.parametrize(
    "dataset", [GAP17, GAP17_LMDB], ids=["gap17", "gap17_lmdb"]
)
def test_filtering(dataset: AtomsDataset):
    assert len(dataset) == 4530

    # filter out all but the first 10 structures
    filtered = dataset.filter_by(config_type="bulk_amo")
    assert len(filtered) == 3410

    bool_idx = [
        structure.info["config_type"] == "bulk_amo" for structure in dataset
    ]
    indexed = dataset[bool_idx]
    assert len(indexed) == 3410
    context = nullcontext() if dataset is GAP17 else pytest.warns(UserWarning)
    with context:
        assert np.all(indexed.info["energy"] == filtered.info["energy"])



from __future__ import annotations

from pathlib import Path
from typing import Iterator

from ase import Atoms
from ase.io import read
from load_atoms.database.backend import BaseImporter, rename, unzip_file
from load_atoms.database.internet import FileDownload
from load_atoms.progress import Progress


class Importer(BaseImporter):
    @classmethod
    def files_to_download(cls) -> list[FileDownload]:
        return [
            FileDownload(
                url="https://zenodo.org/record/7221166/files/data.tar.gz",
                expected_hash="023de5805f15",
            )
        ]

    @classmethod
    def get_structures(
        cls, tmp_dir: Path, progress: Progress
    ) -> Iterator[Atoms]:
        # untar the file
        contents_path = unzip_file(tmp_dir / "data.tar.gz", progress)
        root = contents_path / "data/fig_3"

        table = [
            (
                "loc_run_fig_3a/Final_structures",
                {
                    "config": "continuous random network",
                    "beta": 2.0,
                    "criterion": "local energy",
                },
            ),
            (
                "NN_run_fig_3b/Final_structures_pristine",
                {
                    "config": "paracrystalline",
                    "beta": 2.0,
                    "criterion": "nearest-neighbour energy",
                },
            ),
            (
                "NN_run_fig_3b/Final_structures_thermal",
                {
                    "config": "paracrystalline",
                    "beta": 2.0,
                    "criterion": "nearest-neighbour energy",
                },
            ),
            (
                "loc_run_fig_3c/beta_50/Final_structures",
                {
                    "config": "paracrystalline",
                    "beta": 50.0,
                    "criterion": "local energy",
                },
            ),
            (
                "loc_run_fig_3c/beta_60/Final_structures",
                {
                    "config": "paracrystalline",
                    "beta": 60.0,
                    "criterion": "local energy",
                },
            ),
            (
                "fig_3d/total_beta_0.8/Final_structures",
                {
                    "config": "paracrystalline",
                    "beta": 0.8,
                    "criterion": "total energy",
                },
            ),
        ]

        for path, info in table:
            structures = cls.read_structures(root / path, progress)
            for structure in structures:
                structure.info.update(info)
                yield cls.process_structure(structure)

    @staticmethod
    def read_structures(archive_dir: Path, progress: Progress) -> list[Atoms]:
        extracted = unzip_file(archive_dir / "fin_run.tar.gz", progress)
        return [read(file) for file in sorted(extracted.glob("**/*.xyz"))]  # type: ignore

    @staticmethod
    def process_structure(structure: Atoms) -> Atoms:
        # Remove unwanted arrays
        if "c_1" in structure.arrays:
            del structure.arrays["c_1"]

        # Rotate 90 degrees around the x-axis
        structure.rotate(90, "x", center="COU", rotate_cell=True)

        return rename(
            structure,
            {
                "Forces": "forces",
                "Energy_per_atom": "local_energy",
                "NN_Energy_per_atom": "nn_local_energy",
            },
        )



from __future__ import annotations

from pathlib import Path
from typing import Iterator

import ase.io
from ase import Atoms
from load_atoms.database.backend import BaseImporter, rename, unzip_file
from load_atoms.database.internet import FileDownload
from load_atoms.progress import Progress


class Importer(BaseImporter):
    @classmethod
    def files_to_download(cls) -> list[FileDownload]:
        return [
            FileDownload(
                url="https://zenodo.org/records/10419194/files/database.zip",
                expected_hash="42eb5808b0aa",
            )
        ]

    @classmethod
    def get_structures(
        cls, tmp_dir: Path, progress: Progress
    ) -> Iterator[Atoms]:
        contents = unzip_file(tmp_dir / "database.zip", progress)
        for structure in ase.io.iread(
            contents / "database/training.general_purpose.SiOx.xyz"
        ):
            yield rename(
                structure,
                {
                    "dft_forces": "forces",
                    "dft_energy": "energy",
                    "dft_free_energy": "free_energy",
                    "dft_stress": "stress",
                    "dft_virials": "virial",
                },
            )



import pytest
from load_atoms.utils import (
    LazyMapping,
    freeze_dict,
    generate_checksum,
    intersect,
    k_fold_split,
    lpad,
    matches_checksum,
    random_split,
    split_keeping_ratio,
    union,
)


def test_matches_contents(tmp_path):
    """
    Test that check_file_contents returns True if the contents of a file
    match the given hash.
    """

    fake_file = tmp_path / "file.txt"
    fake_file.write_text("fake file")

    correct_hash = generate_checksum(fake_file)
    incorrect_hash = "0" * 12

    assert matches_checksum(fake_file, correct_hash)
    assert not matches_checksum(fake_file, incorrect_hash)
    assert not matches_checksum(fake_file, correct_hash[:-1])


def test_intersection():
    assert intersect([[1, 2, 3], [2, 3, 4]]) == {2, 3}
    assert intersect(("hi", "hello")) == {"h"}
    assert intersect([]) == set()


def test_union():
    assert union([[1, 2, 3], [2, 3, 4]]) == {1, 2, 3, 4}
    assert union(("hi", "hello")) == {"h", "e", "l", "o", "i"}


def test_lazy_mapping():
    keys = [1, 2, 3]
    _calls = 0

    def loader(key):
        nonlocal _calls
        _calls += 1
        return key + 1

    mapping = LazyMapping(keys, loader)

    with pytest.raises(KeyError):
        mapping[4]

    assert 1 in mapping

    assert mapping[1] == 2
    assert _calls == 1

    mapping[1]
    assert _calls == 1


def test_lpad():
    # check multi-line strings
    s = """\
hello
world"""
    assert (
        lpad(s)
        == """\
    hello
    world"""
    )

    # test custom sep
    assert lpad("hi", 3, fill=":") == ":::hi"


def test_random_split():
    x = list(range(10))

    # test floats
    a, b = random_split(x, [0.5, 0.5])
    assert len(a) == len(b) == 5
    assert len(set(a) & set(b)) == 0, "splits should not overlap"

    # test ints
    a, b = random_split(x, [3, 3])
    assert len(a) == len(b) == 3
    assert len(set(a) & set(b)) == 0, "splits should not overlap"

    # test error
    with pytest.raises(ValueError):
        random_split(x, [10, 10, 10])


def test_k_fold_split():
    x = list(range(10))

    a, b = k_fold_split(x, k=5, fold=0)
    assert a == list(range(8))
    assert b == [8, 9]

    a, b = k_fold_split(x, k=2, fold=1)
    assert a == list(range(5, 10))
    assert b == list(range(5))


def test_split_keeping_ratio():
    things = []
    for i in range(24):
        things.append(
            {
                "id": i,
                "group": 0 if i < 6 else 1 if i < 18 else 2,
            }
        )
    group_ids = [x["group"] for x in things]

    # random split
    a, b, c = split_keeping_ratio(
        things, group_ids, lambda x: random_split(x, [1 / 3, 1 / 3, 1 / 3])
    )
    assert len(a) == len(b) == len(c) == 8

    def proportions(things):
        counts = {0: 0, 1: 0, 2: 0}
        for x in things:
            counts[x["group"]] += 1
        max_count = max(counts.values())
        return {k: v / max_count for k, v in counts.items()}

    assert (
        proportions(a)
        == proportions(b)
        == proportions(c)
        == proportions(things)
    )

    # k-fold split
    x, y = split_keeping_ratio(
        things, group_ids, lambda x: k_fold_split(x, k=3, fold=0)
    )
    assert len(x) == 16
    assert len(y) == 8
    assert proportions(x) == proportions(y) == proportions(things)


def test_frozen_dict():
    d = dict(a=1, b=2, c=3)
    fd = freeze_dict(d)

    with pytest.raises(ValueError, match="read-only"):
        fd["a"] = 3
    assert fd["a"] == 1

    with pytest.raises(ValueError, match="read-only"):
        del fd["a"]
    assert "a" in fd

    with pytest.raises(ValueError, match="read-only"):
        fd.clear()
    assert len(fd) == 3

    with pytest.raises(ValueError, match="read-only"):
        fd.update({"a": 1})
    assert fd["a"] == 1

    with pytest.raises(ValueError, match="read-only"):
        fd.setdefault("a", 1)
    assert fd["a"] == 1

    with pytest.raises(ValueError, match="read-only"):
        fd.pop("a")
    assert fd["a"] == 1

    with pytest.raises(ValueError, match="read-only"):
        fd.popitem()
    assert fd["a"] == 1

    assert list(fd.keys()) == list(d.keys())
    assert list(fd.values()) == list(d.values())
    assert list(fd.items()) == list(d.items())



from load_atoms.database.backend import BASE_GITHUB_URL, SingleFileImporter
from load_atoms.database.internet import FileDownload


class Importer(SingleFileImporter):
    @classmethod
    def file_to_download(cls) -> FileDownload:
        return FileDownload(
            url=f"{BASE_GITHUB_URL}/QM7/QM7.extxyz",
            expected_hash="c9dcec505f4d",
        )



from __future__ import annotations

from io import StringIO
from pathlib import Path
from typing import Iterator

from ase import Atoms
from ase.io import read
from ase.units import Hartree, eV
from load_atoms.database.backend import BaseImporter, unzip_file
from load_atoms.database.internet import FileDownload
from load_atoms.progress import Progress


class Importer(BaseImporter):
    @classmethod
    def files_to_download(cls) -> list[FileDownload]:
        return [
            FileDownload(
                url="https://figshare.com/ndownloader/files/3195389",
                expected_hash="3a63848ac806",
                local_name="qm9.tar.bz2",
            )
        ]

    @classmethod
    def get_structures(
        cls, tmp_dir: Path, progress: Progress
    ) -> Iterator[Atoms]:
        # Unzip the file
        contents_path = unzip_file(tmp_dir / "qm9.tar.bz2", progress)

        # Process each XYZ file
        xyz_files = sorted(contents_path.glob("*.xyz"))
        total_files = len(xyz_files)

        with progress.new_task(
            "Processing QM9 structures", total=total_files
        ) as task:
            for xyz_file in xyz_files:
                yield read_qm9(xyz_file)
                task.update(advance=1)


PROPERTY_KEYS = "index A B C mu alpha homo lumo gap r2 zpve U0 U H G Cv".split()
assert len(PROPERTY_KEYS) == 16
RESCALINGS = [1.0] * 16
for property in "homo lumo gap zpve U0 U H G".split():
    # convert from Hartree to eV
    RESCALINGS[PROPERTY_KEYS.index(property)] = Hartree / eV

# taken from https://figshare.com/ndownloader/files/3195395
OFFSET_COLUMNS = "U0 U H G".split()
OFFSETS = {
    "H": [-0.5002730, -0.4988570, -0.4979120, -0.5109270],
    "C": [-37.846772, -37.845355, -37.844411, -37.861317],
    "N": [-54.583861, -54.582445, -54.581501, -54.598897],
    "O": [-75.064579, -75.063163, -75.062219, -75.079532],
    "F": [-99.718730, -99.717314, -99.716370, -99.733544],
}

# taken from https://figshare.com/ndownloader/files/3195404
BAD_IDS = [
    int(id)
    for id in (Path(__file__).parent.resolve() / "bad_qm9.txt")
    .read_text()
    .splitlines()
]


def read_qm9(file: Path) -> Atoms:
    """
    Read in a single XYZ file from the QM9 dataset, and return it
    as a list of a single Atoms object.

    See the original README for a specification of the format used:
    https://figshare.com/files/3195392
    """
    (
        n,
        property_values,
        *content,
        frequencies,
        smiles,
        inchi,
    ) = file.read_text().replace("*^", "e").splitlines()

    # fake an extxyz file and get ase to read it
    header = 'Properties=species:S:1:pos:R:3:partial_charges:R:1 pbc="F F F"'
    with StringIO("\n".join([n, header, *content])) as extxyz:
        atoms: Atoms = read(extxyz, 0, format="extxyz")  # type: ignore

    # ignore first "gdb" property
    property_values = [float(v) for v in property_values.split()[1:]]
    property_values[0] = int(property_values[0])

    assert len(property_values) == 16
    properties: dict = dict(zip(PROPERTY_KEYS, property_values))

    for name in properties:
        if name in OFFSET_COLUMNS:
            for atom in atoms:
                properties[name] -= OFFSETS[atom.symbol][  # type: ignore
                    OFFSET_COLUMNS.index(name)
                ]
        properties[name] *= RESCALINGS[PROPERTY_KEYS.index(name)]

    properties["frequencies"] = list(map(float, frequencies.split()))

    # molecule characterisation
    properties["smiles"] = smiles.split()[-1]
    properties["inchi"] = inchi.split()[-1]
    properties["geometry"] = property_values[0] not in BAD_IDS

    atoms.info = properties

    return atoms



from __future__ import annotations

import os
from datetime import timedelta
from typing import Protocol

from rich.align import Align
from rich.live import Live
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    ProgressColumn,
    SpinnerColumn,
    TaskID,
    TextColumn,
)
from rich.progress import Progress as RichProgress
from rich.table import Column, Table
from rich.text import Text


class Task(Protocol):
    def update(self, **kwargs):
        ...

    def __enter__(self) -> Task:
        ...

    def __exit__(self, *args):
        ...


class Progress(Protocol):
    def __init__(self, title: str):
        ...

    def new_task(
        self,
        description: str,
        total: int | float | None = None,
    ) -> Task:
        """Add a new task to the progress bar."""
        ...

    def add_text(self, text: str):
        """Add a text row to the progress bar."""

    def refresh(self):
        """Refresh the progress bar."""

    def __enter__(self) -> Progress:
        """Enter the context manager."""
        ...

    def __exit__(self, *args):
        """Exit the context manager."""
        ...


def get_progress_for_dataset(dataset_id: str) -> Progress:
    if os.environ.get("LOAD_ATOMS_VERBOSE") == "0":
        return SilentProgress(dataset_id)
    return VisibleProgress(dataset_id)


class TimeElapsedColumn(ProgressColumn):
    """Renders time elapsed."""

    def render(self, task) -> Text:
        elapsed = task.finished_time if task.finished else task.elapsed
        if elapsed is None:
            return Text("--:--", style="black")
        delta = timedelta(seconds=int(elapsed))
        return Text(":".join(str(delta).split(":")[1:]), style="black")


class PercentColumn(ProgressColumn):
    """Renders percentage complete."""

    def render(self, task) -> Text:
        """Show percentage complete."""
        if task.completed is True:
            return Text("100%", style="black")
        if task.total is None:
            return Text("    ")
        return Text(f"{task.percentage:>3.0f}%", style="black")


class VisibleProgress(Progress):
    def __init__(self, title: str):
        self._progress = RichProgress(
            SpinnerColumn(
                spinner_name="point",
                finished_text="[bold green] ✓ [/bold green]",
                speed=0.5,
                style="black",
            ),
            TextColumn(
                "[progress.description]{task.description}",
                table_column=Column(max_width=40, overflow="fold"),
            ),
            BarColumn(bar_width=20),
            PercentColumn(),
            TimeElapsedColumn(),
        )
        self._table = Table.grid()
        self._table.add_row()
        self._table.add_row(Align(self._progress, align="center"))
        self._live = Live(
            Panel.fit(
                self._table,
                title=f"[bold]{title}",
            ),
            refresh_per_second=10,
            transient=False,
        )

    def new_task(
        self,
        description: str,
        total: int | float | None = None,
    ) -> VisibleTask:
        return VisibleTask(
            self._progress.add_task(description, total=total),
            self._progress,
        )

    def add_text(self, text: str):
        self._table.add_row(Align(Text.from_markup(text), align="center"))

    def __enter__(self):
        self._live.__enter__()
        return self

    def __exit__(self, *args):
        self._progress.refresh()
        self._live.refresh()
        self._live.__exit__(*args)

    def refresh(self):
        self._live.refresh()


class VisibleTask:
    def __init__(self, task_id: TaskID, progress: RichProgress):
        self._task_id = task_id
        self._progress = progress

    def update(self, **kwargs):
        self._progress.update(self._task_id, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._progress.update(self._task_id, completed=True, total=1)


class SilentTask:
    def __init__(self, description: str, total: int | float | None = None):
        ...

    def update(self, **kwargs):
        ...

    def __enter__(self) -> Task:
        return self

    def __exit__(self, *args):
        ...


class SilentProgress(Progress):
    def __init__(self, title: str):
        ...

    def new_task(
        self, description: str, total: int | float | None = None
    ) -> Task:
        return SilentTask(description, total)

    def add_text(self, text: str):
        ...

    def refresh(self):
        ...

    def __enter__(self):
        return self

    def __exit__(self, *args):
        ...



from __future__ import annotations

from pathlib import Path
from typing import Iterator

import h5py
import numpy as np
from ase import Atoms
from load_atoms.database.backend import (
    BaseImporter,
    FileDownload,
)
from load_atoms.progress import Progress

Ha_to_eV = 27.2114079527


class Importer(BaseImporter):
    @classmethod
    def files_to_download(cls) -> list[FileDownload]:
        return [
            FileDownload(
                url="https://springernature.figshare.com/ndownloader/files/18112775",
                expected_hash="fe0ba06198ee",
                local_name="ani1x-release.h5",
            )
        ]

    @classmethod
    def permanent_download_dirname(cls) -> str | None:
        # ensure this file download is persisted so that we don't
        # have to re-download the same file for the ANI-1ccx dataset
        return "ANI"

    @classmethod
    def get_structures(
        cls, tmp_dir: Path, progress: Progress
    ) -> Iterator[Atoms]:
        with h5py.File(tmp_dir / "ani1x-release.h5", "r") as f:
            n_structures = sum(
                data["coordinates"].shape[0] for data in f.values()
            )
            task = progress.new_task(
                "Processing 5M structures",
                total=n_structures,
            )

            # iterate over each chemical formula in the dataset:
            for data in f.values():
                Zs = data["atomic_numbers"]
                coords = data["coordinates"][()]
                dft_energy = data["wb97x_dz.energy"][()]
                dft_dipole = data["wb97x_dz.dipole"][()]
                dft_forces = data["wb97x_dz.forces"][()]
                cc_energy = data["ccsd(t)_cbs.energy"][()]

                for i in range(data["coordinates"].shape[0]):
                    structure = Atoms(positions=coords[i], numbers=Zs)
                    # see: https://www.nature.com/articles/s41597-020-0473-z/tables/2
                    # energy is in hartree, convert to eV
                    structure.info["energy"] = dft_energy[i] * Ha_to_eV
                    # units of e * angstrom
                    structure.info["dipole"] = dft_dipole[i]
                    structure.info["is_in_ccx"] = not np.isnan(cc_energy[i])
                    # forces are in hartree/angstrom, convert to eV/angstrom
                    structure.arrays["forces"] = dft_forces[i] * Ha_to_eV
                    task.update(advance=1)
                    yield structure



"""
The :code:`backend` module is responsible for down/loading datasets by name, 
storing them locally, and serving them to :code:`load-atoms` via the
:func:`~load_atoms.load_dataset` function.
"""

from __future__ import annotations

import shutil
import tempfile
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterator

import ase.io
from ase import Atoms
from typing_extensions import override

from load_atoms.atoms_dataset import (
    AtomsDataset,
    get_file_extension_and_dataset_class,
)
from load_atoms.database.database_entry import (
    LICENSE_URLS,
    DatabaseEntry,
)
from load_atoms.database.internet import FileDownload, download, download_all
from load_atoms.progress import Progress, get_progress_for_dataset
from load_atoms.utils import (
    UnknownDatasetException,
    debug_mode,
    frontend_url,
    testing,
)

BASE_GITHUB_URL = "https://github.com/jla-gardner/load-atoms/raw/main/database"


def load_dataset_by_id(dataset_id: str, root: Path) -> AtomsDataset:
    """
    Load the :class:`AtomsDataset` and corresponding :class:`DatabaseEntry` for
    the given dataset id, saving the dataset to the given ``root`` directory.

    Parameters
    ----------
    name
        The id of the dataset to load.
    root
        The root folder to save the structures to.
    """

    # prepare local paths
    yaml_file_path = root / "database-entries" / f"{dataset_id}.yaml"
    yaml_file_path.parent.mkdir(parents=True, exist_ok=True)

    with get_progress_for_dataset(dataset_id) as progress:
        # down/load the dabase entry for the dataset
        database_entry = get_database_entry(
            dataset_id, yaml_file_path, progress
        )

        # get the file extension and dataset class for the dataset
        extension, dataset_class = get_file_extension_and_dataset_class(
            database_entry.format
        )
        data_file_path = root / f"{dataset_id}.{extension}"

        # if the dataset already exists, load it from disk
        if data_file_path.exists():
            with progress.new_task("Reading from disk"):
                dataset = dataset_class.load(data_file_path)

        # otherwise, use the importer to get the structures
        else:
            # 1. get the Importer class from a suitably down/loaded file
            importer_type: type[BaseImporter] = get_importer_type(
                dataset_id, progress
            )

            # 2. download the files to an appropriate directory
            download_dir_name = importer_type.permanent_download_dirname()
            use_tmp_dir = (
                download_dir_name is None and not debug_mode() and not testing()
            )
            if use_tmp_dir:
                download_dir = Path(tempfile.mkdtemp())
            else:
                download_dir = (
                    root / "raw-downloads" / (download_dir_name or dataset_id)
                )
            download_all(
                importer_type.files_to_download(), download_dir, progress
            )

            # 3. use the importer to get the structures (removing annoying calc)
            def iterator():
                for structure in importer_type.get_structures(
                    download_dir, progress
                ):
                    structure.calc = None
                    yield structure

            try:
                dataset_class.save(data_file_path, iterator(), database_entry)

            except Exception as e:
                # remove the partially created dataset
                if data_file_path.exists():
                    if data_file_path.is_dir():
                        shutil.rmtree(data_file_path)
                    else:
                        data_file_path.unlink()

                raise ValueError(
                    "Failed to import dataset: please report an issue at "
                    "https://github.com/jla-gardner/load-atoms/issues if you "
                    "think this is a bug."
                ) from e

            dataset = dataset_class.load(data_file_path)

            # 4. clean up the temporary directory if necessary
            if use_tmp_dir:
                shutil.rmtree(download_dir)

        # add the usage information to the progress bar
        log_usage_information(database_entry, progress)

        progress.refresh()

    return dataset


class BaseImporter(ABC):
    """
    Base class to inherit from to create new, dataset-specific importers.

    Parameters
    ----------
    files_to_download
        A list of :class:`FileDownload` s
    tmp_dirname
        The name of the temporary directory to download the files to.
    cleanup
        Whether to clean up the temporary directory after processing.
    """

    @classmethod
    def files_to_download(cls) -> list[FileDownload]:
        return []

    @classmethod
    @abstractmethod
    def get_structures(
        cls,
        tmp_dir: Path,
        progress: Progress,
    ) -> Iterator[Atoms]:
        """
        Iterate over :class:`ase.Atoms` objects. All files passed
        to the base class will have already been downloaded
        and verified when this is called.

        Parameters
        ----------
        tmp_dir
            The temporary directory where downloaded files are stored.

        Yields
        ------
        Atoms
            An iterator of ASE Atoms objects processed from the downloaded files
        """

    @classmethod
    def permanent_download_dirname(cls) -> str | None:
        """
        Get a path to the directory where the files should be saved.
        If ``None`` (the default), is returned, the files will be downloaded
        to a temporary directory, and removed after the dataset is imported.
        """
        return None


class SingleFileImporter(BaseImporter):
    @classmethod
    @abstractmethod
    def file_to_download(cls) -> FileDownload:
        ...

    @classmethod
    def files_to_download(cls) -> list[FileDownload]:
        return [cls.file_to_download()]

    @override
    @classmethod
    def get_structures(
        cls, tmp_dir: Path, progress: Progress
    ) -> Iterator[Atoms]:
        file_path = tmp_dir / Path(cls.files_to_download()[0].local_name)
        with progress.new_task(f"Reading {file_path.resolve()}"):
            for atoms in cls._read_file(file_path):
                yield cls.process_atoms(atoms)

    @classmethod
    def process_atoms(cls, atoms: Atoms) -> Atoms:
        return atoms

    @classmethod
    def _read_file(cls, file_path: Path) -> Iterator[Atoms]:
        yield from ase.io.iread(file_path, index=":")


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~ HELPERS ~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


def get_database_entry(
    dataset_id: str,
    yaml_file_path: Path,
    progress: Progress,
) -> DatabaseEntry:
    from load_atoms import __version__ as load_atoms_version

    if not yaml_file_path.exists():
        try:
            download(
                DatabaseEntry.remote_url_for_yaml(dataset_id),
                yaml_file_path,
                progress,
            )
        except Exception as e:
            raise UnknownDatasetException(dataset_id) from e

    db_entry = DatabaseEntry.from_yaml_file(yaml_file_path)

    if (
        db_entry.minimum_load_atoms_version is not None
        and db_entry.minimum_load_atoms_version > load_atoms_version
    ):
        raise Exception(
            f"Dataset {dataset_id} requires load-atoms version "
            f">={db_entry.minimum_load_atoms_version} "
            f"(current version: {load_atoms_version}). "
            "Please upgrade load-atoms to load this dataset "
            "(e.g. `pip install --upgrade load-atoms`)."
        )

    return db_entry


def get_importer_type(
    dataset_id: str,
    progress: Progress,
) -> type[BaseImporter]:
    importer_name = DatabaseEntry.importer_file_stem(dataset_id)
    expected_importer_path = (
        Path(__file__).parent / "importers" / f"{importer_name}.py"
    )

    if not expected_importer_path.exists():
        try:
            download(
                DatabaseEntry.remote_url_for_importer(dataset_id),
                expected_importer_path,
                progress,
            )
        except Exception as e:
            # couldn't download the importer:
            raise UnknownDatasetException(dataset_id) from e

    try:
        return __import__(
            f"load_atoms.database.importers.{importer_name}",
            fromlist=["Importer"],
        ).Importer

    except Exception as e:
        raise Exception(
            f"Unable to load dataset {dataset_id} due to a problem loading "
            "the dataset's importer file. Please try updating load-atoms:\n"
            "  pip install --upgrade load-atoms\n"
            "If the problem persists, please report an issue at:\n"
            "  https://github.com/jla-gardner/load-atoms/issues"
        ) from e


def log_usage_information(info: DatabaseEntry, progress: Progress):
    progress.add_text("\n")

    name = f"[bold]{info.name}[/bold]"
    if info.license is not None:
        style = f"dodger_blue2 link={LICENSE_URLS[info.license]} underline"
        progress.add_text(
            f"The {name} dataset is covered by the "
            f"[{style}]{info.license}[/] license."
        )
    if info.citation is not None:
        progress.add_text(
            f"Please cite the {name} dataset " "if you use it in your work."
        )
    progress.add_text(f"For more information about the {name} dataset, visit:")
    url = frontend_url(info)
    url_style = f"dodger_blue2 underline link={url}"
    progress.add_text(f"[{url_style}]load-atoms/{info.name}")


def unzip_file(file_path: Path, progress: Progress) -> Path:
    """Unzip a file and return the path to the extracted directory.

    Parameters
    ----------
    file_path
        The path to the file to unzip.
    progress
        A :class:`Progress` object to track the unzip progress.
    """

    extract_to = file_path.parent / f"{file_path.name}-extracted"
    if not extract_to.exists():
        with progress.new_task(
            f"Unzipping {file_path.resolve()}",
        ):
            shutil.unpack_archive(file_path, extract_dir=extract_to)
    return extract_to


def rename(atoms: Atoms, mapping: dict[str, str]) -> Atoms:
    """Rename the properties of an Atoms object."""

    for old_name, new_name in mapping.items():
        if old_name in atoms.arrays:
            atoms.arrays[new_name] = atoms.arrays.pop(old_name)
        elif old_name in atoms.info:
            atoms.info[new_name] = atoms.info.pop(old_name)
    return atoms
